import hudson.model.*
import jenkins.model.*
import groovy.json.JsonSlurper

def execute_command(cmd, args, verbose=false) {
    if (!cmd.toString()?.trim()) {
        cmd = "python "+System.getenv("JENKINS_HOME")+"/jenkins-scripts/common/aws_ec2.py"
    }
    Process process = "${cmd} ${args}".execute()
    def out = new StringBuffer()
    def err = new StringBuffer()
    process.consumeProcessOutput( out, err )
    process.waitFor()
    if (verbose == true) {
        println "\n<<<< START CMD: ${args.split()[0]} >>>>"
        println "OUT:\n" + out.toString()
        println "ERR:\n" + err.toString()
        println "<<<< END CMD >>>>\n"
    }

    def HashMap ret_items
    if (err.toString()?.trim()) {
        println "You got error message:"
        println out.toString()
        println err.toString()
        assert false
    } else {
        out_str = out.toString().replace("None", "''")
        if (out_str.contains("-----BEGIN RC OUTPUT-----") && \
            out_str.contains("-----END RC OUTPUT-----")) {
            ret_data = out_str.substring( \
                           out_str.indexOf("-----BEGIN RC OUTPUT-----") + 26, \
                           out_str.indexOf("-----END RC OUTPUT-----"))
            def jsonSlurper = new JsonSlurper()
            ret_items = jsonSlurper.parseText(ret_data.replaceAll("'","\""))
        }
    }
    return ret_items
}

class WorkerConf {
    protected Properties conf

    WorkerConf() {
        conf = new Properties()
        conf.load(new FileInputStream(System.getenv('JENKINS_HOME') \
                                      + '/init.groovy.d/setup.properties'))
    }

    Integer max_slaves() { conf.EC2_INSTANCE_CAP_STR.toInteger() }
    Integer executors_per_slave() { conf.EC2_NUMBER_OF_EXECUTORS.toInteger() }
    String name_prefix() { conf.EC2_DESCRIPTION }
    String keyname() { conf.AWS_CLOUD_KEYNAME }

    String ami_id() { conf.EC2_AMI_ID }
    String availability_zone() { conf.EC2_AV_ZONE }
    String placement() {
        return "Tenancy:default," \
             + "GroupName:," \
             + "AvailabilityZone:${this.availability_zone()}"
    }
    String security_groups() { conf.EC2_SECURITY_GROUPS.replaceAll(" ","") }
    String remote_fs() { conf.EC2_REMOTE_FS }
    String instance_type() { conf.EC2_INSTANCE_TYPE }
    String labels() { conf.EC2_LABEL_STRING }
    String remote_user() { conf.EC2_REMOTE_ADMIN }
    String ssh_port() { conf.EC2_SSH_PORT }
    String subnet_id() { conf.EC2_SUBNET_ID }
    String launch_timeout() { conf.EC2_LAUNCH_TIMEOUT }
    long idle_termination_minutes() { conf.EC2_IDLE_TERMINATION_MINUTES.toInteger() }
    String credential_id() { conf.EC2_CREDENTIAL_ID }
    String tag_name() { conf.EC2_TAG_NAME }
    String tag_hostname() { conf.EC2_TAG_HOSTNAME }
    String tag_env() { conf.EC2_TAG_ENV }
    String tag_source() { conf.EC2_TAG_SOURCE }
    String tags() {
        return "Name:${this.tag_name()}," \
             + "hostname:${this.tag_hostname()}," \
             + "env:${this.tag_env()}," \
             + "source:${this.tag_source()}," \
             + "jenkins_slave_type:demand_${this.name_prefix()}"
    }
}

class WorkerConf_JENKINS_IMAGER extends WorkerConf {
    @Override
    String name_prefix() { conf.EC2_WORKER_IMAGER_DESCRIPTION }
    String tag_source() { conf.EC2_WORKER_IMAGER_TAG_SOURCE }

    List check_queue_list() { conf.EC2_WORKER_IMAGER_QUEUE_CHECK_LIST.split(",") }
}

class WorkerConf_OBS_WORKER_NORMAL extends WorkerConf {
    @Override
    String name_prefix() { conf.EC2_WORKER_OBS_NORMAL_DESCRIPTION }
}

class WorkerConf_OBS_WORKER_POWER extends WorkerConf {
    @Override
    String name_prefix() { conf.EC2_WORKER_OBS_POWER_DESCRIPTION }
}

class SlaveStatus {
    private Map slave_stat = ['CURR_NUMBER_OF_NODES': 0,
                              'CURR_TOTAL_EXECUTORS': 0,
                              'CURR_BUSY_EXECUTORS': 0,
                              'CURR_IDLE_EXECUTORS': 0]
    private Map slave_info = [:]

    SlaveStatus(String name_prefix) {
        for (slave in Hudson.instance.slaves) {
            def slaveComputer = slave.getComputer()
            if ("${slaveComputer.getName()}".startsWith(name_prefix)) {
                slave_stat['CURR_NUMBER_OF_NODES']++
                slave_stat['CURR_TOTAL_EXECUTORS'] += slaveComputer.countExecutors()
                slave_stat['CURR_BUSY_EXECUTORS'] += slaveComputer.countBusy()
                if (!slaveComputer.offline && slaveComputer.isAlive()) {
                    slave_stat['CURR_IDLE_EXECUTORS'] += slaveComputer.countIdle()
                }
                def this_slave = ["name":slaveComputer.getName(),
                                  "idle_since":System.currentTimeMillis() \
                                               - slaveComputer.getIdleStartMilliseconds(),
                                  "host":slave.getLauncher().getHost(),
                                  "object":slaveComputer]
                slave_info[slaveComputer.getName()] = this_slave
            }
        }
    }

    Integer number_of_nodes() {
        slave_stat['CURR_NUMBER_OF_NODES']
    }
    Integer total_executors() {
        slave_stat['CURR_TOTAL_EXECUTORS']
    }
    Integer idle_executors() {
        slave_stat['CURR_IDLE_EXECUTORS']
    }
    Integer busy_executors() {
        slave_stat['CURR_BUSY_EXECUTORS']
    }
    Map get_slave_info() {
        return slave_info
    }
    void disconnect_node(slaveName) {
        if (slaveName in slave_info) {
            slave_info[slaveName]["object"].setTemporarilyOffline(true, null)
            slave_info[slaveName]["object"].disconnect(null)
        }
    }
    void delete_node(slaveName) {
        if (slaveName in slave_info) {
            slave_info[slaveName]["object"].doDoDelete()
        }
    }
}

def get_aws_status(worker_conf) {
    args = "ping -t ${worker_conf.tags()}"
    return execute_command("", args, verbose=true)
}

def terminate_aws_ec2_instances(instance_ids) {
    args = "terminate -i ${instance_ids.join(',')}"
    return execute_command("", args, verbose=true)
}

def create_aws_ec2_instances(ami_id,
                             min_count,
                             max_count,
                             keyname,
                             security_group_ids,
                             instance_type,
                             placement,
                             subnet_id,
                             ebsoptimized,
                             tags) {

    args = "create --amiid ${ami_id} " \
           + " --keyname ${keyname} " \
           + " --mincount ${min_count} --maxcount ${max_count} " \
           + " --securitygroupids ${security_group_ids} " \
           + " --instancetype ${instance_type} " \
           + " --placement ${placement} " \
           + " --subnetid ${subnet_id} " \
           + " --ebsoptimized ${ebsoptimized} " \
           + " --tags ${tags}"
    return execute_command("", args, verbose=true)
}

def get_worker_conf(purpose) {

    def worker_conf = Class.forName("WorkerConf_${purpose}", true, \
                                    this.class.classLoader).newInstance()
    assert worker_conf
    println "\nWORKER CONFIGURATION (${purpose})"
    println "  MAX_SLAVES: ${worker_conf.max_slaves()}"
    println "  EXECUTORS_PER_SLAVE: ${worker_conf.executors_per_slave()}"
    println "  NAME_PREFIX: ${worker_conf.name_prefix()}\n"
    return worker_conf
}

//TODO: FIXME:
e = { filepath ->
    evaluate(new File(System.getenv("JENKINS_HOME") + "/init.groovy.d/" + filepath))
}
create_slave_node = e("Module_Node")

def create_slaves(worker_conf, num_nodes_to_create) {
    ec2s = create_aws_ec2_instances(ami_id=worker_conf.ami_id(),
                                    min_count=1,
                                    max_count=num_nodes_to_create,
                                    keyname=worker_conf.keyname(),
                                    security_group_ids=worker_conf.security_groups(),
                                    instance_type=worker_conf.instance_type(),
                                    placement=worker_conf.placement(),
                                    subnet_id=worker_conf.subnet_id(),
                                    ebsoptimized='False',
                                    tags=worker_conf.tags())

    Date date = new Date()
    String curr_date = date.format("yyyyMMdd.HHmmss")

    ec2s.each{ k, v ->
        println "Creating jenkins node ${worker_conf.name_prefix()} (${v.instance_id}) at " + curr_date
        println "  Instance ID: ${v.instance_id}"
        println "  IP Address: ${v.private_ip_address}"
        println "  Launch Time: ${v.launch_time}"
        create_slave_node(
            instance = Jenkins.getInstance(),
            name = "${worker_conf.name_prefix()} (${v.instance_id})",
            remoteFS = "${worker_conf.remote_fs()}",
            numExecutors = "${worker_conf.executors_per_slave()}",
            labelString = "${worker_conf.labels()}",
            sshHost = "${v.private_ip_address}",
            sshPort = "${worker_conf.ssh_port()}",
            sshCredentials = "${worker_conf.credential_id()}",
            userId = "${worker_conf.remote_user()}",
            description = "${v.launch_time}"
        )
        instance.save()
    }
    println ""
    return ec2s
}

Integer how_many_slaves_to_create(worker_conf, num_requested_executors) {

    // Check slave nodes are free enough to allocate image jobs
    def slave_stat = new SlaveStatus(worker_conf.name_prefix())
    println "\nSLAVE STATUS (${worker_conf.name_prefix()}"
    println "  NUMBER_OF_NODES: ${slave_stat.number_of_nodes()}"
    println "  TOTAL EXECUTORS: ${slave_stat.total_executors()}"
    println "  IDLE EXECUTORS: ${slave_stat.idle_executors()}"
    println ""
    slave_stat.get_slave_info().each{ k, v ->
        println "    ${k}:${v}"
    }
    println ""
    def Integer final_nodes_to_create = 0
    if (slave_stat.number_of_nodes() >= worker_conf.max_slaves() ) {
        println "REACHED MAXIMUM SLAVES"
        return 0
    }
    if (slave_stat.idle_executors() >= num_requested_executors) {
        println "HAVE ENOUGH EXECUTORS"
        return 0
    }

    println "CALCULATING..."
    def Integer min_to_create = \
        Math.min(num_requested_executors - slave_stat.idle_executors(), \
                  (worker_conf.max_slaves() - slave_stat.number_of_nodes()) \
                  * worker_conf.executors_per_slave())
    def Double nodes_to_create = min_to_create.div(worker_conf.executors_per_slave())
    final_nodes_to_create = Math.min(Math.ceil(nodes_to_create).intValue(), \
                             worker_conf.max_slaves() - slave_stat.number_of_nodes())
    println "MIN(" + num_requested_executors + "-" + slave_stat.idle_executors() \
            + ", (" + worker_conf.max_slaves() + "-" + slave_stat.number_of_nodes() \
            + ")*" + worker_conf.executors_per_slave() + ")" \
            + " = " + min_to_create
    println "FINAL NUMBER OF SLAVES TO CREATE: ${final_nodes_to_create} \n"
    return final_nodes_to_create
}

def worker_ondemand_create_request(worker_conf, Integer num_requested_executors) {

    println "YOU REQUESTED ${num_requested_executors} executors!"
    get_aws_status(worker_conf)

    // Check slave nodes are free enough to allocate image jobs
    def num_nodes_to_create = how_many_slaves_to_create(worker_conf, num_requested_executors)

    if (num_nodes_to_create > 0) {
        inst_info = create_slaves(worker_conf, num_nodes_to_create)
        println "\"TitleDisplay\": \"Create(${inst_info.size()}/${num_requested_executors})\""
    }
}

def worker_ondemand_revoke_request(worker_conf) {

    def bRevoke = true

    // Check slave nodes for running builds
    def slave_stat = new SlaveStatus(worker_conf.name_prefix())
    if (slave_stat.busy_executors() > 0) {
        println "\nYOU HAVE BUSY EXECUTORS."
        return
    }

    // Check build queue
    Jenkins.instance.queue.items.find { it ->
        if (it.task.name in worker_conf.check_queue_list()) {
            println "\nPENDING BUILD IN THE QUEUE (${it.task.name}:${it.id}). STOP REVOKE OPERATION."
            bRevoke = false
            return true
        }
    }
    if (bRevoke != true) { return }

    // Check build
    worker_conf.check_queue_list().find { it ->
        def item = hudson.model.Hudson.instance.getItem(it)
        if (item.isInQueue() || item.isBuilding()) {
            println "YOU HAVE QUEUED OR BUILDING ITEMS FOR ${it}"
            bRevoke = false
            return true
        }
    }
    if (bRevoke != true) { return }

    existing_instance_ids = []
    get_aws_status(worker_conf).each { k, v ->
        if (v["state"] != "terminated") {
            existing_instance_ids.add(k)
        }
    }
    println "\nEXISTING INSTANCE INFO:\n${existing_instance_ids}\n"

    Date date = new Date()
    String curr_date = date.format("yyyyMMdd.HHmmss")

    existing_slave_ids = []
    inst_for_terminate = []
    slave_stat.get_slave_info().each { k, v ->
        def matcher = v["name"] =~ /^${worker_conf.name_prefix()} \((i-[0-9a-f]+)\)$/
        if (matcher?.matches()) {
            def inst_id = matcher.group(1)
            existing_slave_ids.add(inst_id)
            if ( v["idle_since"].toLong().div(1000*60) >= worker_conf.idle_termination_minutes()) {
                println ".... Delete candidate ${k} : " + v["idle_since"].toLong().div(1000*60)
                // Mark the node as offline and disconnect
                slave_stat.disconnect_node(v["name"])
                if (inst_id in existing_instance_ids) {
                    inst_for_terminate.add(inst_id)
                }
            } else {
                println ".... TIMER NOT EXPIRED FOR ${k} : " + v["idle_since"].toLong().div(1000*60)
            }
        } else {
            println "SLAVE \"${k}\" IS NOT MY CHILD."
        }
    }
    println "\nEXISTING NODE INFO:\n${existing_slave_ids}\n"

    println inst_for_terminate
    if (inst_for_terminate) {
        terminate_aws_ec2_instances(inst_for_terminate)
        println "\"TitleDisplay\": \"Delete(${inst_for_terminate.size()})\""
    }

    slave_stat.get_slave_info().each { k, v ->
        def matcher = v["name"] =~ /^${worker_conf.name_prefix()} \((i-[0-9a-f]+)\)$/
        if (matcher?.matches()) {
            def inst_id = matcher.group(1)
            // Delete the node
            if (inst_id in inst_for_terminate) {
                println "Deleting jenkins node ${k} at " + curr_date \
                    + " Cause: " + v["idle_since"].toLong().div(1000*60)
                slave_stat.delete_node(v["name"])
            }
        }
    }

    dangled_jenkins_nodes = existing_slave_ids.toSet() - existing_instance_ids.toSet()
    if (dangled_jenkins_nodes) {
        println "\nDangled jenkins nodes:\n${dangled_jenkins_nodes}"
    }

    dangled_ec2_instances = existing_instance_ids.toSet() - existing_slave_ids.toSet()
    if (dangled_ec2_instances) {
        println "\nDangled EC2 instances:\n${dangled_ec2_instances}"
        assert false
    }
}

def __main__() {

    def buildEnv = build.getEnvironment(listener)

    if (buildEnv["ONDEMAND_SLAVE_CONFIGURATION_ENABLED"] != "1") {
        println "FEATURE NOT ENABLED"
        return
    }

    action = buildEnv['ACTION']
    purpose = buildEnv['PURPOSE']
    requested_num_executors = buildEnv['REQUESTED_NUM_EXECUTORS']

    assert (buildEnv['PURPOSE'] == 'JENKINS_IMAGER')

    def worker_conf = get_worker_conf(purpose)

    if (action == "REQUEST_WORKER") {
        worker_ondemand_create_request(worker_conf, \
                                     requested_num_executors.toInteger())
    } else if(action == "REVOKE_WORKER") {
        worker_ondemand_revoke_request(worker_conf)
    } else {
        println "Invalid action: " + action
    }
}

__main__()

return 0

